package rabbit.cache;

// $Id: NCache.java,v 1.26 2005/08/03 17:00:41 robo Exp $

import java.io.*;
import java.util.*;
import java.util.zip.*;

import rabbit.http.*;
import rabbit.util.*;

/** The NCache is like a Map in lookup/insert/delete
 *  The NCache is persistent over sessions (saves itself to disk).
 *  The NCache is selfcleaning, that is it removes old stuff.
 */
public class NCache implements Runnable 
{
    private static String DIR = "cache";              // standard dir.
    private static long SIZE  = 10 * 1024 * 1024;     // 10 MB.
    private static long CT = 24 * 60 * 60 * 1000;     // 1 day.

    private static String CACHEINDEX = "cache.index"; // the indexfile.
    private static String TEMPDIR = "temp";
    private static int filesperdir = 256;             // reasonable?

    private boolean changed = false;                  // have we changed?
    private Thread cleaner = null;                    // remover of old stuff.
    private int cleanLoopTime = 60 * 1000;     // how long the cleaner sleeps between loops.

    private long maxSize = 0;
    private long cacheTime = 0;
    private long fileNo = 0;
    private long currentSize = 0;
    private String dir = null;
    private Map<FiledKey, NCacheEntry> htab = null; // mapping Object => CacheEntry.
    private List<NCacheEntry> vec = null;

    private File tempdir = null;
    private Object dirLock = new Object ();

    /** Create a cache that uses default values.
     */
    public NCache () {
	this (DIR, SIZE, CT);
    }

    /** Create a cache for the given directory.
     * @param dir the name of the directory where the cache should store its files.
     */
    public NCache (String dir) {
	this (dir, SIZE, CT);
    }
    
    /** Create a cache for the given directory with given parameters.
     * @param dir the name of the directory where the cache should store its files.
     * @param maxSize the maximum size of the proxy.
     */
    public NCache (String dir, long maxSize) {
	this (dir, maxSize, CT);
    }
    
    /** Create a cachehandler.
     * @param dir the basedir for the cachehandler.
     * @param maxSize the maximum size of the proxy.
     * @param cacheTime how long the proxy should keep things unless specified otherwise.
     */
    public NCache (String dir, long maxSize, long cacheTime) {
	htab = new HashMap<FiledKey, NCacheEntry> ();
	vec = new ArrayList<NCacheEntry> ();
	setMaxSize (maxSize);
	setCacheTime (cacheTime);
	setCacheDir (dir);
	cleaner = new Thread (this, getClass ().getName () + ".cleaner");
	cleaner.setDaemon (true);
	cleaner.start ();
    } 


    /** Get the name of the directory where the cache stores its files.
     * @return the name of the cachedir
     */
    public synchronized String getCacheDir () {
	return dir;
    }
    
    /** Sets the cachedir. This will flush the cache and make 
     *  it try to read in the cache from the new dir.
     * @param newDir the name of the new directory to use.
     */
    public synchronized void setCacheDir (String newDir) {
	// save old cachedir.
	if (dir != null)
	    writeCacheIndex ();  
	
	// does new dir exist?
	dir = newDir;
	File dirtest = new File (dir);
	if (!dirtest.exists ()) {
	    dirtest.mkdirs ();
	    if (!dirtest.exists ()) {
		System.err.println ("couldnt create cachedir, exiting");
		System.exit (-1);
	    }
	} else if (dirtest.isFile ()) {
	    System.err.println ("Cachedir is a file, that is bad, exiting");
	    System.exit (-1);
	}
	
	synchronized (dirLock) {
	    tempdir = new File (dir + File.separator + TEMPDIR);
	    if (!tempdir.exists ()) {
		tempdir.mkdir ();
		if (!tempdir.exists ()) {
		    System.err.println ("couldlnt create cache tempdir, exiting");
		    System.exit (-1);
		}
	    } 
	    else if (tempdir.isFile ()) {
		System.err.println ("Cache temp dir is a file, that is bad, exiting");
		System.exit (-1);
	    }
	}
	// move to new dir.
	readCacheIndex ();
    }

    /** Get the maximum size for this cache.
     * @return the maximum size in bytes this cache.
     */
    public long getMaxSize () {
	return maxSize;
    }
    
    /** Set the maximum size for this cache.
     * @param newMaxSize the new maximum size for the cache.
     */
    public void setMaxSize (long newMaxSize) {
	maxSize = newMaxSize;
    }

    /** Get the number of miliseconds the cache stores things usually.
     *  This is the standard expiretime for objects, but you can set it for 
     *  CacheEntries individially if you want to.
     *  NOTE 1: dont trust that an object will be in the cache this long.
     *  NOTE 2: dont trust that an object will be removed from the cache when it expires.
     * @return the number of miliseconds objects are stored normally.
     */
    public long getCacheTime () {
	return cacheTime;
    }
    
    /** Set the standard expiry-time for CacheEntries
     * @param newCacheTime the number of miliseconds to keep objects normally.
     */
    public void setCacheTime (long newCacheTime) {
	cacheTime = newCacheTime;
    }

    /** Get how long time the cleaner sleeps between cleanups.
     */
    public int getCleanLoopTime () {
	return cleanLoopTime;
    }

    /** Set how long time the cleaner sleeps between cleanups.
     * @param newCleanLoopTime the number of miliseconds to sleep. 
     */
    public void setCleanLoopTime (int newCleanLoopTime) {
	cleanLoopTime = newCleanLoopTime;
    }
    
    /** Get the current size of the cache
     * @return the current size of the cache in bytes.
     */
    public synchronized long getCurrentSize () {
	return currentSize;
    }

    /** Get the current number of entries in the cache.
     * @return the current number of entries in the cache.
     */
    public long getNumberOfEntries () {
	return htab.size ();
    }
    
    /** Check that the key really exists. 
     *  This may cause Filed data to be read in. 
     */
    private boolean checkKey (NCacheEntry e) {
	return (e.getKey () != null);
    }

    /** Check that the data hook exists. 
     */
    private boolean checkHook (NCacheEntry e) {
	Object hook = e.getRealDataHook ();
	if (hook != null && hook instanceof FiledHook) {
	    String entryName = getEntryName (e);
	    File f = new File (entryName + ".hook");
	    if (!f.exists ())
		return false;
	}
	// no hook is legal.
	return true;
    }

    /** Get the CacheEntry assosiated with given object.
     * @param o the key.
     * @return the NCacheEntry or null (if not found).
     */ 
    public NCacheEntry getEntry (Object o) {
	NCacheEntry ent;
	synchronized (this) {
	    ent = htab.get (new MemoryKey (o));
	}
	if (ent != null) {
	    /* used to check "!checkKey (ent) &&" but that is unneccessary
	     * get will read in data.
	     */
	    if (!checkHook (ent)) {
		// bad entry...
		remove (ent);
	    }
	}
	/*if (ent != null)
	  ent.setVisited (new Date ());*/
	return ent;
    }

    /** Get the file name for a real cache entry. */
    public String getEntryName (NCacheEntry ent) {
	return getEntryName (ent.getId (), true);
    }

    /** Get the file name for a cache entry. 
     * @param id the id of the cache entry
     * @param real false if this is a temporary cache file, true if it is a realized entry.
     */
    public String getEntryName (long id, boolean real) {
	StringBuilder sb = new StringBuilder (50);
	sb.append (dir);
	sb.append (File.separator);
	if (!real) {
	    sb.append (TEMPDIR);
	} else {
	    long fdir = id / filesperdir;
	    sb.append (fdir);
	}
	sb.append (File.separator);
	sb.append (id);
	return sb.toString ();
    }


    /** Reserve space for a CacheEntry with key o.
     * @param o the key for the NCacheEntry.
     * @return a new CacheEntry initialized for the cache.
     */
    public NCacheEntry newEntry (Object o) {
	long newId = 0;
	// allocate the id for the new entry.
	synchronized (this) {
	    newId = fileNo;
	    fileNo++;
	}
	// allocate the entry.
	NCacheEntry entry = new NCacheEntry (o, newId);
	long exp = System.currentTimeMillis () + getCacheTime ();
	entry.setExpires (exp);
	return entry;
    }

    /** Insert a CacheEntry into the cache.
     * @param ent the CacheEntry to store.
     */
    public void addEntry (NCacheEntry ent) {
	if (ent == null)
	    return;	
	File cfile = new File (getEntryName (ent.getId (), false)); 
	if (!cfile.exists()) {
	    return;
	}

	long fdir = ent.getId () / filesperdir;
	File f = new File (dir, "" + fdir);
	String newName = getEntryName (ent.getId (), true); 
	File nFile = new File (newName);
	synchronized (dirLock) {
	    if (f.exists ()) {
		if (f.isFile ()) {
		    System.err.println ("Doh, wanted cachedir is a file!");
		    System.exit (-1);
		}
		// good situation...
	    } else {
		f.mkdir ();
	    }	
	    boolean bool = cfile.renameTo (nFile);
	}
	cfile = new File (newName);	
	ent.setSize (cfile.length ());
	ent.setCacheTime (new Date ());

	Object hook = ent.getDataHook (this);
	if (hook != null) {
	    FiledHook fh = new FiledHook ();
	    fh.storeHook (this, ent, hook);
	    ent.setDataHook (fh);
	}
	
	Object realKey = ent.getKey ();
	FiledKey fk = new FiledKey ();
	fk.storeKey (this, ent, realKey);
	synchronized (this) {
	    ent.setKey (fk);
	    remove (realKey);
	    htab.put (fk, ent);
	    currentSize += ent.getSize ();
	    vec.add (ent);
	}	
	
	changed = true;
    }

    /** Signal that a cache entry have changed.
     */
    public void entryChanged (NCacheEntry ent) {
	changed = true;	
    }

    private void removeHook (String base, String extension) {
	String hookName = base + extension;
	// remove possible hook before file...
	File hfile = new File (hookName);
	if (hfile.exists ())
	    hfile.delete ();
    }
    
    /** Remove the Entry with key o from the cache.
     * @param o the key for the CacheEntry.
     */
    public void remove (Object o) {
	NCacheEntry rent;
	synchronized (this) {
	    if (o == null) // Odd, but seems to happen. 		
		return;    // Probably removed by someone else before enumeration gets to it.
	    FiledKey fk = new MemoryKey (o);
	    rent = htab.get (fk);
	    if (rent != null) {
		// remove entries while it is still in htab.
		vec.remove (rent);
		currentSize -= rent.getSize ();
		htab.remove (fk);
	    }
	}
		
	if (rent != null) {
	    // this removes the key => htab.remove can not work..
	    String entryName = getEntryName (rent);
	    removeHook (entryName, ".hook");
	    removeHook (entryName, ".key");
	    rent.setKey (null);
	    rent.setDataHook (null);
	    File cfile = new File (entryName);
	    if (cfile.exists ()) {
		File p = cfile.getParentFile ();
		cfile.delete ();
		// humm, until NT does rename in a nice manner check for tempdir.
		synchronized (dirLock) {
		    if (p.exists () && !p.equals (tempdir)) {		    
			String ls[] = p.list ();
			if (ls != null && ls.length == 0)
			    p.delete();
		    }
		}
	    }
	}
    }

    /** Clear the Cache from files. 
     */
    public synchronized void clear () {
	for (Object o : new ArrayList<Object> (htab.keySet ()))
	    remove (o);
	vec.clear (); // just to be safe.
	currentSize = 0;
	changed = true;
    }

    /** Get the CacheEntries in the cache.
     *  Note! some entries may be invalid if you have a corruct cache.
     * @return an Enumeration of the CacheEntries.
     */    
    public Collection<NCacheEntry> getEntries () {
	return htab.values ();
    }

    /** Read the info from an old cache.
     */
    private void readCacheIndex () {
	fileNo = 0;
	currentSize = 0;
	htab = new HashMap<FiledKey, NCacheEntry> ();
	vec = new ArrayList<NCacheEntry> ();
	try {
	    String name = dir + File.separator + CACHEINDEX;
	    ObjectInputStream is = 
		new ObjectInputStream (new GZIPInputStream (new FileInputStream (name)));
	    fileNo = is.readLong ();
	    currentSize = is.readLong ();
	    int size = is.readInt ();
	    Map<FiledKey, NCacheEntry> hh = 
		new HashMap<FiledKey, NCacheEntry> ((int)(size * 1.2));
	    for (int i = 0; i < size; i++) {
		FiledKey fk = (FiledKey)is.readObject ();
		fk.setCache (this);
		NCacheEntry entry = (NCacheEntry)is.readObject ();
		entry.setKey (fk);
		hh.put (fk, entry);
	    }
	    htab = hh;
	    vec = (List<NCacheEntry>)is.readObject ();
	    is.close ();
	} catch (IOException e) {
	    /*
	    System.err.println ("Couldnt read " + 
				dir + File.separator + CACHEINDEX + 
				", This is bad( but not serius).\n" + 
				"Treating as empty. " + e);
	    */
	} catch (ClassNotFoundException e) {
	    System.err.println ("Couldn't find classes? " + e);
	}
    }
    
    /** Make sure that the cache is written to the disk.
     */
    public void flush () {
	writeCacheIndex ();
    }

    /** Store the cache to disk so we can reuse it later.
     */
    private void writeCacheIndex () {
	try {
	    String name = dir + File.separator + CACHEINDEX;
	    
	    ObjectOutputStream os = 
		new ObjectOutputStream (new GZIPOutputStream (new FileOutputStream (name)));
	    
	    synchronized (this) {
		os.writeLong (fileNo);
		os.writeLong (currentSize);
		os.writeInt (htab.size ());
		for (Map.Entry<FiledKey, NCacheEntry> me : htab.entrySet ()) {
		    os.writeObject (me.getKey ());
		    os.writeObject (me.getValue ());		    
		}
		os.writeObject (vec);
	    }
	    os.close ();
	} catch (IOException e) {
	    System.err.println ("Couldnt write " + dir + File.separator + 
				CACHEINDEX + ", This is serious!\n" + e);
	    System.exit (-1);
	}	
    }

    /** Loop in a cleaning loop.
     */
    public void run () {
	Thread.currentThread ().setPriority (Thread. MIN_PRIORITY);
	while (true) {
	    try {
		Thread.sleep (cleanLoopTime);
	    } catch (InterruptedException e) {
		//System.err.println ("Cache interrupted");
	    }
	    
	    // actually for a busy cache this will lag...
	    // but I dont care for now... 
	    long milis = System.currentTimeMillis ();
	    Collection<NCacheEntry> entries;
	    synchronized (this) {
		entries = new ArrayList<NCacheEntry> (htab.values ());
	    }
	    for (NCacheEntry ce : entries) {
		long exp = ce.getExpires ();
		if (exp < milis) 
		    remove (ce.getKey ());
	    }

	    // IF SIZE IS TO BIG REMOVE A RANDOM AMOUNT OF OBJECTS.
	    // What we have to be careful about: we must not remove the same
	    // elements two times in a row, this method remove the "oldest" in 
	    // a sense.

	    if (getCurrentSize () > getMaxSize ())
		changed = true;
	    while (getCurrentSize () > getMaxSize ()) {
		synchronized (this) {
		    remove (vec.get (0).getKey ());
		}
	    }
	    
	    if (changed) {
		//System.out.println ("writing " + dir + File.separator + CACHEINDEX + "!");
		writeCacheIndex ();
		changed = false;
	    }
	}
    }

    /** Configure the cache system from the given config.
     * @param config the properties describing the cache settings.
     * @throws IllegalConfigurationException if some setting is strange.
     */
    public void setup (SProperties config) throws IllegalConfigurationException {
	if (config == null)
	    return;
	String cachedir = config.getProperty ("directory", "/tmp/rabbit/cache/");
	setCacheDir (cachedir);
	String cmsize = config.getProperty ("maxsize", "10");
	try {
	    setMaxSize (Long.parseLong (cmsize) * 1024 * 1024);     // in MB
	} catch (NumberFormatException e) { 
	    throw (new IllegalConfigurationException ("Bad number for cache maxsize: '" + cmsize + "'"));
	}
	String ctime = config.getProperty ("cachetime", "24");
	try {
	    setCacheTime (Long.parseLong (ctime) * 1000 * 60 * 60); // in hours.
	} catch (NumberFormatException e) { 
	    throw (new IllegalConfigurationException ("Bad number for cache cachetime: '" + ctime + "'"));
	}
	String cleanTime = config.getProperty ("cleanloop", "60");
	try {
	    setCleanLoopTime (Integer.parseInt (cleanTime) * 1000); // in seconds.
	} catch (NumberFormatException e) { 
	    throw (new IllegalConfigurationException ("Bad number for cache cleanloop: '" + cleanTime + "'"));
	}
    }
}
