package rabbit.filter;

// $Id: AccessFilter.java,v 1.10 2005/08/02 20:50:21 robo Exp $

import java.io.*;
import java.net.*;
import java.util.*;

import rabbit.proxy.*;
import rabbit.util.*;

/** This is a class that filters access based on ip adress.
 */
public class AccessFilter implements IPAccessFilter {
    private String accessfile;
    private List<IPAccess> iplist = new ArrayList<IPAccess> ();
    private static final String DEFAULTCONFIG = "conf/access";

    /** Filter based on a socket.
     * @param s the Socket to check.
     * @return true if the Socket should be allowed, false otherwise.
     */
    public boolean doIPFiltering (Socket s) {
	int isize =  iplist.size ();
	for (int i = 0; i < isize; i++) {
	    IPAccess ipa = iplist.get (i);
	    if (ipa.inrange (s.getInetAddress ()))
		return true;
	}
	return false;
    }

    /** Setup this class.
     * @param properties the Properties to get the settings from.
     */
    public void setup (Logger logger, SProperties properties) {
	String file = properties.getProperty ("accessfile", DEFAULTCONFIG);
	loadAccess (logger, file);
    }
     
    /** Read the data (accesslists) from a file.
     * @param filename the name of the file to read from.
     */
    private void loadAccess (Logger logger, String filename) {
	filename = filename.replace ('/', File.separatorChar);
	accessfile = filename;
	
	FileReader fr = null;
	try {
	    fr = new FileReader (accessfile);
	    loadAccess (logger, fr);	    
	} catch (IOException e) {
	    logger.logError (Logger.FATAL, "Accessfile '" + accessfile + 
			    "' not found:(allowing no one is a bad idea) " + e);
	    System.exit (-1);
	} finally {
	    if (fr != null) {
		try {
		    fr.close ();
		} catch (IOException e) {
		    logger.logError (Logger.FATAL, "failed to close accessfile: " + e);		    
		}
	    }
	}
    }
    
    /** Loads in the accessess allowed from the given Reader
     * @param r the Reader were data is available
     */
    public void loadAccess (Logger logger, Reader r) throws IOException {
	List<IPAccess> iplist = new ArrayList<IPAccess> ();
	LineNumberReader br = new LineNumberReader (r);
	String line = null;
	while ((line = br.readLine ()) != null) {
	    // remove comments....
	    int index = line.indexOf ('#');
	    if (index >= 0) 
		line = line.substring (0,index);
	    if (line.equals (""))
		continue;
	    StringTokenizer st = new StringTokenizer (line);
	    if (st.countTokens () != 2) {
		logger.logError (Logger.WARN, "Bad line in accessconf:" + br.getLineNumber());
		continue;
	    }
	    String low = st.nextToken ();
	    InetAddress lowip = null;
	    try {
		lowip = InetAddress.getByName (low);	    
	    } catch (UnknownHostException e) {
		logger.logError (Logger.WARN, "Bad host: " + low + " at line:" + br.getLineNumber());
		continue;	    
	    }
	    
	    String high = st.nextToken ();
	    InetAddress highip = null;
	    try {
		highip = InetAddress.getByName (high);
	    } catch (UnknownHostException e) {
		logger.logError (Logger.WARN, "Bad host: " + high + " at line:" + br.getLineNumber());
		continue;	    
	    }
	    
	    if (lowip != null && highip != null) 
		iplist.add (new IPAccess (lowip, highip));
	}
	br.close ();
	this.iplist = iplist;
    }

    /** Saves the accesslist from the given Reader.
     * @param r the Reader with the users.
     */
    public void saveAccess (Logger logger, Reader r) throws IOException {
	if (accessfile == null) 
	    return;
	BufferedReader br = new BufferedReader (r);
	PrintWriter fw = new PrintWriter (new FileWriter (accessfile));
	String line;
	while ((line = br.readLine ()) != null)
	    fw.println (line);
	fw.flush ();
	fw.close ();
	br.close ();
    }

    /** Get the list of ips allowed
     */
    public List<IPAccess> getIPList () {
	return iplist;
    }
}
