package rabbit.filter;

// $Id: HTTPBaseFilter.java,v 1.39 2005/08/02 00:06:38 robo Exp $

import java.io.*;
import java.math.*;
import java.net.*;
import java.util.*;

import rabbit.http.*;
import rabbit.proxy.*;
import rabbit.proxy.Proxy;
import rabbit.util.*;

/** This is a class that filter http headers to make them nice.
 *  This filter sets up username and password if supplied and 
 *  also sets up keepalive.
 */
public class HTTPBaseFilter implements HTTPFilter {
    public static final String NOPROXY = "http://noproxy.";
    private static final BigInteger ZERO = new BigInteger ("0");
    private static final BigInteger ONE = new BigInteger ("1");

    private String userfile = null;
    private List<String> removes = new ArrayList<String> ();
    private Map<String, String> users = new HashMap<String, String> ();
    private boolean cookieId = false;

    /** 
     */
    public HTTPBaseFilter () {
    }

    /** We got a proxy authentication, handle it...
     * @param uap the authentication string.
     * @param con the Connection.
     */
    private void handleProxyAuthentication (String uap, Connection con) {
	// guess we should handle digest here also.. :-/
	if (uap.startsWith ("Basic ")) {
	    uap = uap.substring ("Basic ".length ());
	    String userapass = Coder.uudecode (uap);
	    int i = -1;
	    if ((i = userapass.indexOf (":")) > -1) {
		String userid = userapass.substring (0, i);
		String pass = userapass.substring (i+1);
		con.setUserName (userid);
		con.setPassWord (pass);
	    }
	}
    }

    /** Handle the authentications.
     *  If we have a proxy-authentication we set the connections username and password.
     *  We also rewrite authentications in the URL to a standard header,
     *  since java does not handle them.
     * @param header the Request.
     * @param con the Connection.
     */
    private void handleAuthentications (HTTPHeader header, Connection con) {
	String uap = header.getHeader ("Proxy-Authorization");
	if (uap != null) 
	    handleProxyAuthentication (uap, con);
	
        /*
         * Java URL:s doesn't handle user/pass in the URL as in rfc1738:
         * //<user>:<password>@<host>:<port>/<url-path>
         *
         * Convert these to an Authorization header and remove from URI.
         */	
        String requestURI = header.getRequestURI();
        
        int s3, s4, s5;
        if ((s3 = requestURI.indexOf("//")) >= 0 &&
            (s4 = requestURI.indexOf('/', s3 + 2)) >= 0 &&
            (s5 = requestURI.indexOf('@', s3 + 2)) >= 0 &&
            s5 < s4) {
            
            String userPass = requestURI.substring(s3 + 2, s5);
            header.setHeader("Authorization", "Basic " +
			     Coder.uuencode(userPass));
	    
            header.setRequestURI(requestURI.substring(0, s3 + 2) +
				 requestURI.substring(s5 + 1));
        }
    }

    /** Check if this is a noproxy request, and if so handle it.
     * @param requri the requested resource.
     * @param header the actual request.
     * @param con the Connection.
     * @return the new request URI
     */
    private String handleNoProxyRequest (String requri, HTTPHeader header, Connection con) {
	requri = "http://" + requri.substring (NOPROXY.length ());
	header.setRequestURI (requri);
	con.setMayUseCache (false);
	con.setMayCache (false);
	con.setMayFilter (false);
	return requri;
    }

    /** Check that the requested URL is valid and if it is a meta request.
     * @param requri the requested resource.
     * @param header the actual request.
     * @param con the Connection.
     */
    private HTTPHeader handleURLSetup (String requri, HTTPHeader header, Connection con) {
	try {
	    // is this request to our self?
	    Proxy proxy = con.getProxy ();
	    if (requri != null && requri.charAt (0) == '/') {
		requri = 
		    "http://" + proxy.getHost ().getHostName () + 
		    ":" + proxy.getPort () + requri;
		header.setRequestURI (requri);
	    }
	    URL url = new URL (requri);
	    header.setHeader ("Host",  url.getPort () > -1 ? 
			      url.getHost () + ":" + url.getPort () : url.getHost ());
	    int urlport = url.getPort ();
	    // this gives a DNS-error if no DNS is available, and since we have not
	    // decided if we should proxy it up the chain yet, do string comparison..
	    //InetAddress urlhost = InetAddress.getByName (url.getHost ());	    
	    String uhost = url.getHost ();
	    if (proxy.isSelf (uhost, urlport)) {
		con.setMayUseCache (false);
		con.setMayCache (false);
		con.setMayFilter (false);
		if (!isValidUser (con.getUserName (), con.getPassWord ()) && 
		    !isPublic (url)) {
		    HTTPHeader ret = 
			con.getResponseHandler ().get407 (uhost + ":" + urlport, url);
		    return ret;
		}
		con.setMeta (true);
	    }
	} catch (MalformedURLException e) {   
	    return con.getResponseHandler ().get400 (e);
	}
	return null;
    }
    
    /** Remove all "Connection" tokens from the header.
     * @param header the HTTPHeader that needs to be cleaned.
     */
    private void removeConnectionTokens (HTTPHeader header) {
	List cons = header.getHeaders ("Connection");	
	for (int i = 0; i < cons.size (); i++) {
	    String val = (String)cons.get (i);
	    /* ok, split it... */
	    int s = -1;
	    int start = 0;
	    while (start < val.length ()) {
		while (val.length () > start + 1 && (val.charAt (start) == ' ' ||
						     val.charAt (start) == ','))
		    start++;
		if (val.length () > start + 1 && val.charAt (start) == '"') {
		    start++;
		    s = val.indexOf ('"', start);
		    while (s >= -1 && val.charAt (s - 1) == '\\' && val.length () > s + 1)
			s = val.indexOf ('"', s + 1);			
		    if (s == -1)
			s = val.length ();
		    String t = val.substring (start, s).trim ();
		    /* ok, unquote the value... */
		    StringBuilder sb = new StringBuilder (t.length ());
		    for (int c = 0; c < t.length (); c++) {
			char z = t.charAt (c);
			if (z != '\\')
			    sb.append (z);
		    }
		    t = sb.toString ();
		    header.removeHeader (t);
		    s = val.indexOf (',', s + 1);
		    if (s == -1)
			start = val.length ();
		    else 
			start = s + 1;
		} else {
		    s = val.indexOf (',', start + 1);
		    if (s == -1)
			s = val.length ();
		    String t = val.substring (start, s).trim ();
		    header.removeHeader (t);
		    start = s + 1;
		}
	    }
	}
    }

    private HTTPHeader checkMaxForwards (Connection con, HTTPHeader header, String val) {
	try {
	    BigInteger bi = new BigInteger (val);
	    if (bi.equals (ZERO)) {
		if (header.getMethod ().equals ("TRACE")) {
		    HTTPHeader ret = con.getResponseHandler ().get200 ();
		    ret.setContent (header.toString ());
		    return ret;
		} else {
		    HTTPHeader ret = con.getResponseHandler ().get200 ();
		    ret.setHeader ("Allow", "GET,HEAD,POST,OPTIONS,TRACE");
		    ret.setHeader ("Content-Length", "0");
		    return ret;
		}
	    } else {
		BigInteger b3 = bi.subtract (ONE);
		header.setHeader ("Max-Forwards", b3.toString ());
	    }
	} catch (NumberFormatException e) {
	    con.getProxy ().logError (Logger.WARN, "Bad number for Max-Forwards: '" + val + "'");
	}
	return null;
    }

    /** test if a socket/header combination is valid or return a new HTTPHeader.
     * @param socket the Socket that made the request.
     * @param header the actual request made.
     * @param con the Connection handling the request.
     * @return null if everything is fine or a HTTPHeader describing the error (like a 403).
     */
    public HTTPHeader doHTTPInFiltering (Socket socket, HTTPHeader header, Connection con) {
	// ok, no real header then dont do a thing.
	if (header.isDot9Request ()) {
	    con.setMayCache (false);
	    con.setMayUseCache (false);
	    con.setKeepalive (false);
	    return null;
	}
	
	handleAuthentications (header, con);
	
	boolean maychunk = true;
	boolean mayKeepAlive = true;
    
	String requestVersion = header.getHTTPVersion ().toUpperCase ();
	if (requestVersion.equals ("HTTP/1.1")) {
	    String host = header.getHeader ("Host");
	    if (host == null) {
		Exception exe = new Exception ("No host header set in HTTP/1.1 request");
		HTTPHeader ret = 
		    con.getResponseHandler ().get400 (exe);
		return ret;
	    }
	    maychunk = true;    
	    String closeit = header.getHeader ("Proxy-Connection");
	    if (closeit == null)
		closeit = header.getHeader ("Connection");
	    mayKeepAlive = (closeit == null || !closeit.equalsIgnoreCase ("close"));
	} else {
	    header.setHTTPVersion ("HTTP/1.1");
	    maychunk = false;
	    // stupid netscape to not follow the standards, only "Connection" should be used...
	    String keepalive = header.getHeader ("Proxy-Connection");
	    mayKeepAlive = (keepalive != null && keepalive.equalsIgnoreCase ("Keep-Alive"));
	    if (!mayKeepAlive) {
		keepalive = header.getHeader ("Connection");
		mayKeepAlive = (keepalive != null && keepalive.equalsIgnoreCase ("Keep-Alive"));
	    }	
	}
	
	boolean useCached = true;
	boolean cacheAllowed = true;
	// damn how many system that use cookies with id's
	/*System.out.println ("auth: " + header.getHeader ("authorization") + 
			    ", cookie:" + header.getHeader ("cookie") + 
			    ", Pragma: " + header.getHeader ("Pragma") + 
			    ", Cache: " + header.getHeader ("Cache-Control")); */
	//String cached = header.getHeader ("Pragma");
	List ccs = header.getHeaders ("Cache-Control");
	for (int i = 0; i < ccs.size (); i++) {
	    String cached = (String)ccs.get (i);
	    cached = cached.trim ();
	    if (cached.equals ("no-store")) {
		useCached = false;
		cacheAllowed = false;
	    } else if (cached.equals ("no-cache")) {
		useCached = false;
	    } else if (cached.equals ("no-transform")) {
		useCached = false;     // cache is transformed.
		cacheAllowed = false;  // dont store, no point.
		con.setMayFilter (false);
	    }
	}

	ccs = header.getHeaders ("Pragma");
	for (int i = 0; i < ccs.size (); i++) {
	    String cached = (String)ccs.get (i);
	    cached = cached.trim ();
	    if (cached.equals ("no-cache")) {
		useCached = false;
	    }
	}

	
	String method = header.getMethod ().trim ();
	if (!method.equals ("GET") && !method.equals ("HEAD")) {
	    useCached = false;
	    cacheAllowed = false;
	    //mayKeepAlive = false;
	} else if (method.equals ("HEAD")) {
	    maychunk = false;
	}
	con.setChunking (maychunk);

	String mf = header.getHeader ("Max-Forwards");
	if (mf != null) {
	    HTTPHeader ret = checkMaxForwards (con, header, mf);
	    if (ret != null) {
		return ret;
	    }
	}

	String auths = header.getHeader ("authorization");
	if (auths != null) {
	    useCached = false;     // dont use cached files, 
	    cacheAllowed = false;  // and dont cache it.
	} else if (cookieId) {
	    String cookie = header.getHeader ("cookie");
	    String lccookie = null;
	    if (cookie != null &&     // cookie-passwords suck.
		(((lccookie = cookie.toLowerCase ()).indexOf ("password") >= 0) ||
		 (lccookie.indexOf ("id") >= 0))) {
		useCached = false;     // dont use cached files, 
		cacheAllowed = false;  // and dont cache it.
	    }
	}
	con.setMayUseCache (useCached);
	con.setMayCache (cacheAllowed);
	con.setKeepalive (mayKeepAlive);
	
	String requri = header.getRequestURI ();
	if (requri.toLowerCase ().startsWith (NOPROXY)) 
	    requri = handleNoProxyRequest (requri, header, con);       
		
	HTTPHeader headerr = handleURLSetup (requri, header, con);
	if (headerr != null) 
	    return headerr;
	
	removeConnectionTokens (header);
	int rsize = removes.size ();
	for (int i = 0; i < rsize; i++) {
	    String r = removes.get (i);
	    header.removeHeader (r);
	}
	
	Proxy proxy = con.getProxy ();
	if (proxy.isProxyConnected ()) {
	    String auth = proxy.getProxyAuthString ();
	    // it should look like this (using RabbIT:RabbIT):
	    // Proxy-authorization: Basic UmFiYklUOlJhYmJJVA==
	    header.setHeader ("Proxy-authorization", "Basic " + Coder.uuencode (auth));
	}

	// try to use keepalive backwards.
	// This is not needed since it is a HTTP/1.1 request.
	// header.setHeader ("Connection", "Keep-Alive");

	return null;
    }

    private boolean checkCacheControl (String cachecontrol) {
	StringTokenizer st = new StringTokenizer (cachecontrol, ",");
	while (st.hasMoreTokens ()) {
	    String cached = st.nextToken ().trim ();
	    if (cached.equals ("no-store"))
		return false;
	    if (cached.equals ("private")) 
		return false;
	}
	return true;
    }

    /** test if a socket/header combination is valid or return a new HTTPHeader.
     * @param socket the Socket that made the request.
     * @param header the actual request made.
     * @param con the Connection handling the request.
     * @return null if everything is fine or a HTTPHeader describing the error (like a 403).
     */
    public HTTPHeader doHTTPOutFiltering (Socket socket, HTTPHeader header, Connection con) {
	boolean useCache = true;
	//String cached = header.getHeader ("Pragma");
	List ccs = header.getHeaders ("Cache-Control");
	for (int i = 0; i < ccs.size (); i++) {
	    String cached = (String)ccs.get (i);
	    if (cached != null) 
		useCache &= checkCacheControl (cached);
	}
	
	String status = header.getStatusCode ().trim ();
	if (!(status.equals ("200") || status.equals ("206") || status.equals ("304"))) {
	    con.setKeepalive (false);    
	    useCache = false;
	}

	String age = header.getHeader ("Age");
	long secs = 0;
	if (age == null)
	    age = "0";
	try {
	    secs = Long.parseLong (age);
	} catch (NumberFormatException e) {
	    // ignore, we already have a warning for this..
	}
	if (secs > 60 * 60 * 24) 
	    header.setHeader ("Warning", "113 RabbIT \"Heuristic expiration\"");
	
	header.setResponseHTTPVersion ("HTTP/1.1");
	con.setMayCache (useCache);

	/** Dont filter compressed pages... someone else has already thought about speed... */
	String ce = header.getHeader ("Content-Encoding");
	if (ce != null &&
	    (ce.equalsIgnoreCase ("gzip") 
	     || (ce.equalsIgnoreCase ("compress"))))
	    con.setMayFilter (false);

	/** Try to make sure that IE can handle NTLM authentication. */
	/** This does not work. */
	/*
	List ls = header.getHeaders ("WWW-Authenticate");
	for (Iterator i = ls.iterator (); i.hasNext (); ) {
	    String s = (String)i.next ();
	    if (s.indexOf ("Negotiate") != -1 || 
		s.indexOf ("NTLM") != -1) {
		con.setMayFilter (false);
		con.setChunking (false);
	    }
	}
	*/
	
	removeConnectionTokens (header);
	int rsize = removes.size ();
	for (int i = 0; i < rsize; i++) {
	    String r = removes.get (i);
	    header.removeHeader (r);
	}
	
	String d = header.getHeader ("Date");
	if (d == null) {
	    // ok, maybe we should check if there is an Age set
	    // otherwise we can do like this.
	    header.setHeader ("Date", HTTPDateParser.getDateString (new Date ()));
	}
	
	String cl = header.getHeader ("Content-Length");
	if (cl == null) 
	    if (!con.getChunking ())
		con.setKeepalive (false);

	return null;
    }

    /** Setup this class with the given properties.
     * @param properties the new configuration of this class.
     */
    public void setup (Logger logger, SProperties properties) {
	removes.clear ();
	String rs = properties.getProperty ("remove", "");
	StringTokenizer st = new StringTokenizer (rs, ",");
	while (st.hasMoreElements ()) {
	    String r = st.nextToken ().trim ();
	    removes.add (r);
	}
	userfile = properties.getProperty ("userfile", "conf/users");
	FileReader fr = null;
	try {
	    fr = new FileReader (userfile);
	    users = loadUsers (fr);
	} catch (FileNotFoundException e) {
	    logger.logError (Logger.WARN, "could not load the users file: '" + userfile + "'. " + e);
	} catch (IOException e) {
	    logger.logError (Logger.WARN, "Error while loading the users file: '" + userfile + "'. " + e);	    
	} finally {
	    if (fr != null) {
		try {
		    fr.close ();
		} catch (IOException e) {
		    logger.logError (Logger.WARN, "Error when closing file: " + e);
		}
	    }
	}
	String cid = properties.getProperty ("cookieid", "false");
	cookieId = cid.equals ("true");
    }
    
    /** Load the users from the given Reader.
     * @param r the Reader with the users.
     */
    public static Map<String, String> loadUsers (Reader r) 
	throws IOException {
	BufferedReader br = new BufferedReader (r);
	String line;
	Map<String, String> u = new HashMap<String, String> ();
	while ((line = br.readLine ()) != null) {
	    StringTokenizer st = new StringTokenizer (line, ": \n\t");
	    String name = "";
	    String pass = "";
	    if (st.hasMoreTokens ())
		name = st.nextToken ();
	    else 
		continue;
	    if (st.hasMoreTokens ())
		pass = st.nextToken ();
	    u.put (name, pass);
	}
	return u;
    }

    /** Saves the users from the given Reader.
     * @param r the Reader with the users.
     */
    public void saveUsers (Reader r) throws IOException {
	if (userfile == null) 
	    return;
	BufferedReader br = new BufferedReader (r);
	PrintWriter fw = new PrintWriter (new FileWriter (userfile));
	String line;
	while ((line = br.readLine ()) != null) 
	    fw.println (line);
	fw.flush ();
	fw.close ();
    }

    /** Return the hash of users.
     */
    public Map<String, String> getUsers () {
	return users;
    }

    /** Return the hash of users.
     */
    public void setUsers (Map<String, String> users) {
	this.users = users;
    }
    
    /** Check if a user/password combination is valid.
     * @param username the username.
     * @param password the decrypted password.
     * @return true if both username and password match a valid user.
     */
    public boolean isValidUser (String username, String password) {
	if (username == null)
	    return false;
	String pass = users.get (username);
	return (pass != null && password != null && pass.equals (password));	
    }

    /** Check if a given url is a public URL of the Proxy.
     * @param url the URL to check.
     * @return true if this url has public access, false otherwise.
     */
    public boolean isPublic (URL url) {
	String file = url.getFile ();
	if (file.startsWith ("/FileSender/public/"))
	    return true;
	return false;
    }
}
