package rabbit.filter;

// $Id: SQLProxyAuth.java,v 1.6 2005/08/02 20:50:21 robo Exp $

import java.net.MalformedURLException;
import java.net.Socket;
import java.net.URL;
import java.sql.*;
import rabbit.http.HTTPHeader;
import rabbit.proxy.Connection;
import rabbit.util.Logger;
import rabbit.util.SProperties;

/** This is a filter that requires users to use 
 *  proxy-authentication, using users in a sql table.
 *
 *  Will read the following parameters from the config file: 
 *  <ul>
 *  <li>driver 
 *  <li>url
 *  <li>user
 *  <li>password
 *  <li>select
 *  </ul>
 */
public class SQLProxyAuth implements HTTPFilter {
    private String url = null;
    private String dbuser = null;
    private String dbpwd = null;
    private String select = null;
    private String DEFAULT_SELECT = 
    "select password from users where username = ?";

    private java.sql.Connection db = null;

    public SQLProxyAuth (){
    }

    private synchronized void initConnection () throws SQLException {
	db = DriverManager.getConnection (url, dbuser, dbpwd);
    }

    /** test if a socket/header combination is valid or return a new HTTPHeader.
     *  Check that the user has been authenticate..
     * 
     *  Check if we want to cache user info...
     * 
     * @param socket the Socket that made the request.
     * @param header the actual request made.
     * @param con the Connection handling the request.
     * @return null if everything is fine or a HTTPHeader describing the error (like a 403).
     */
    public HTTPHeader doHTTPInFiltering (Socket socket, HTTPHeader header, Connection con) {
	if (con.getMeta ())
	    return null;
	String username = con.getUserName ();
	String pwd = con.getPassWord ();
	if (username == null || pwd == null) 
	    return getError (con, header);
	String rpwd = null;
	try { 
	    rpwd = getBackendPassword (con.getProxy (), username);
	} catch (SQLException e) {
	    con.getProxy ().logError (Logger.WARN, 
				      "Exception when trying to get user: " + e);
	    closeDB (con);
	}
	if (rpwd == null || !rpwd.equals (pwd))
	    return getError (con, header);
	return null;
    }

    /** test if a socket/header combination is valid or return a new HTTPHeader.
     * @param socket the Socket that made the request.
     * @param header the actual request made.
     * @param con the Connection handling the request.
     * @return null if everything is fine or a HTTPHeader describing the error (like a 403).
     */
    public HTTPHeader doHTTPOutFiltering (Socket socket, HTTPHeader header, Connection con) {
	return null;
    }

    private HTTPHeader getError (Connection con, HTTPHeader header) {
	try {
	    return con.getResponseHandler ().get407 ("internet", 
						     new URL (header.getRequestURI ()));
	} catch (MalformedURLException e) {
	    con.getProxy ().logError (Logger.WARN, "Bad url: " + e);
	    return con.getResponseHandler ().get407 ("internet", null);
	}
    }

    private synchronized void closeDB (Connection con) {
	if (db == null)
	    return;
	try {
	    db.close ();
	} catch (SQLException e) {
	    con.getProxy ().logError (Logger.WARN, "failed to close database: " + e);
	}
	db = null;
    } 

    private String getBackendPassword (Logger logger, String username) 
	throws SQLException {
	if (db == null)
	    initConnection ();
	PreparedStatement ps = null;
	ResultSet rs = null;
	try {
	    ps = db.prepareStatement (select);
	    ps.setString (1, username);
	    rs = ps.executeQuery ();
	    if (rs.next ()) {
		return rs.getString (1);		
	    }
	    return null;
	} finally {
	    if (rs != null) {
		try {
		    rs.close ();
		} catch (SQLException e) {
		    logger.logError (Logger.ERROR, "failed to close resultset: " + e);
		}
	    }
	    if (ps != null) {
		try {
		    ps.close ();
		} catch (SQLException e) {
		    logger.logError (Logger.ERROR, "failed to close statement: " + e);
		}		
	    }
	} 
    }

    /** Setup this class with the given properties.
     * @param logger the Logger to output errors/warnings on.
     * @param properties the new configuration of this class.
     */
    public void setup (Logger logger, SProperties properties) {
	String driver = properties.getProperty ("driver");
	try {
	    Class.forName (driver);
	} catch (ClassNotFoundException e) {
	    logger.logError (Logger.ERROR, "Failed to load driver");
	}
	url = properties.getProperty ("url");
	dbuser = properties.getProperty ("user");
	dbpwd = properties.getProperty ("password");
	select = properties.getProperty ("select", DEFAULT_SELECT);
    }    
}
