package rabbit.handler;

// $Id: ImageHandler.java,v 1.24 2005/08/02 20:50:21 robo Exp $

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;

import rabbit.cache.NCache;
import rabbit.http.HTTPHeader;
import rabbit.io.HTTPInputStream;
import rabbit.io.MultiOutputStream;
import rabbit.proxy.Connection;
import rabbit.proxy.Proxy;
import rabbit.util.Logger;
import rabbit.util.SProperties;

/** This class filters images to low quality jpegs.
 */
public class ImageHandler extends BaseHandler {
    private static final String STD_CONVERT = "/usr/X11R6/bin/convert";
    private static final String STD_CONVERT_ARGS = "-quality 5 $filename[0] jpeg:$filename.c";
    private SProperties config = new SProperties ();
    private boolean doConvert = true;

    private boolean converted = false;
    private long origSize = -1;
    private long lowQualitySize = -1;
    private File convertedFile = null;
    private boolean notEnoughContent = false;

    // For creating the factory.
    public ImageHandler () {	
    }
    
    /** Create a new ImageHansler for the given request.
     * @param con the Connection handling the request.
     * @param request the actual request made.
     * @param response the actual response.
     * @param contentstream the stream to read data from.
     * @param clientstream the stream to write data to.
     * @param maycache May we cache this request? 
     * @param mayfilter May we filter this request?
     * @param size the size of the data beeing handled.
     */
    public ImageHandler (Connection con, 
			 HTTPHeader request, 
			 HTTPHeader response, 
			 HTTPInputStream contentstream, 
			 MultiOutputStream clientstream,
			 boolean maycache,
			 boolean mayfilter,
			 long size) {
	super (con, request, response, contentstream, clientstream, 
	       maycache, mayfilter, size);
	if (size == -1)
	    con.setKeepalive (false);
	con.setChunking (false);
    }

    @Override public Handler getNewInstance (Connection connection,
					     HTTPHeader header, 
					     HTTPHeader webheader, 
					     HTTPInputStream contentStream, 
					     MultiOutputStream out, 
					     boolean maycache, 
					     boolean mayfilter, 
					     long size) {
	ImageHandler ih = new ImageHandler (connection, header, webheader, 
					    contentStream, out, maycache, 
					    mayfilter, size);
	ih.config = config;
	ih.doConvert = doConvert;
	return ih;
    }
    
    /** Images needs to be cacheable to be compressed.
     * @return true
     */
    protected boolean mayCacheFromSize () {
	return true;
    }

    /** Check if this handler may force the cached resource to be less than the cache max size.
     * @return false
     */
    protected boolean mayRestrictCacheSize () {
	return false;
    }
    
    /** Read in the image
     * @throws IOException if reading of the image fails.
     */
    protected void readImage () throws IOException {
	byte[] v = new byte[2048];
	int i;
	long total = 0;
	try {
	    while ((size < 0 || total < size) && (i = contentstream.read(v)) > 0) {
		total += i;
		cacheStream.write (v, 0, i);
	    }
	} finally {
	    if (size > 0 && size != total) 
		setPartialContent (total, size);
	    cacheStream.close ();
	    NCache cache = con.getProxy ().getCache ();
	    origSize = new File (cache.getEntryName (entry.getId (), false)).length ();
	}
    }

    private void closeStreams (Process ps) throws IOException {
	ps.getInputStream ().close ();
	ps.getOutputStream ().close ();
	ps.getErrorStream ().close ();
    }
    
    /** Convert the image into a low quality jpeg.
     * @throws IOException if conversion fails.
     */
    protected void convertImage () throws IOException {
	String convert = config.getProperty ("convert", STD_CONVERT);
	String convargs = config.getProperty ("convertargs", STD_CONVERT_ARGS);
	
	int idx = 0;
	Proxy proxy = con.getProxy ();
	String entryName = proxy.getCache ().getEntryName (entry.getId (), false);
	try {
	    while ((idx = convargs.indexOf ("$filename")) > -1) {
		convargs = convargs.substring (0, idx) + entryName + 
		    convargs.substring (idx + "$filename".length());
	    }
	    String command = convert + " " + convargs;
	    
	    Process ps = Runtime.getRuntime ().exec (command);
	    try {
		ps.waitFor ();
		closeStreams (ps);
		int exitValue = ps.exitValue (); 
		if (exitValue != 0)
		    logError (Logger.WARN, 
			      "Bad conversion: " + entryName + 
			      ", got exit value: " + exitValue);
	    } catch (InterruptedException e) {
		logError (Logger.WARN, 
			  "Interupted during wait for: " + entryName);
	    }
	    
	    convertedFile = new File (entryName + ".c");
	    lowQualitySize = convertedFile.length ();
	    if (origSize > lowQualitySize) {
		response.setHeader ("Content-Type", "image/jpeg");
		/** We need to remove the existing file first for 
		 *  windows system, they will not overwrite files in a move.
		 *  Spotted by: Michael Mlivoncic 
		 */
		File oldEntry = new File (entryName);
		oldEntry.delete (); 
		if (convertedFile.renameTo (new File (entryName)))
		    convertedFile = null;
	    }
	} finally { 
	    if (convertedFile != null) {
		convertedFile.delete ();
	    }
	}
	size = (lowQualitySize < origSize ? lowQualitySize : origSize);
	response.setHeader ("Content-length", "" + size);
	con.setExtraInfo ("imageratio:" + origSize + "/" + lowQualitySize + 
			  "=" + ((float)lowQualitySize / origSize));
	
	contentstream.close ();
	contentstream = new HTTPInputStream (new FileInputStream (entryName), this);
	convertedFile = null;
    }

    /** Try to convert the image. This is done like this:
     *  <xmp>
     *  super.addCacheStream ();
     *  readImage();
     *  convertImage();
     *  clientstream.removeOutputStream (cacheStream);
     *  </xmp>
     *  We have to use the cachefile to convert the image, and if we
     *  convert it we dont want to write the file to the cache later
     *  on. 
     */
    protected void tryconvert () throws IOException {
	if (doConvert && mayfilter && maycache) {
	    if (size > 2000) {
		super.addCacheStream ();
		// check if cache setup worked.
		if (cacheStream == null)
		    return;
		readImage ();
		convertImage ();
		converted = true;
		clientstream.removeOutputStream (cacheStream);
		cacheStream = null;
	    }	
	} 
    }
    
    /** Write the response header
     * @throws IOException if writing the response fails.
     */
    protected void writeHeader () throws IOException {
	tryconvert ();
	super.writeHeader ();
    }

    /** Set up the cache stream if available.
     * @throws IOException if a cachestream couldnt be set up.
     */
    public void addCacheStream () throws IOException {
	if (!converted)
	    super.addCacheStream ();
    }

    /** clear up the mess we made (remove intermediate files etc).
     * @throws IOException if cleaning fails.
     */
    protected void finish () throws IOException {
	if (convertedFile != null)
	    convertedFile.delete ();
	if (notEnoughContent)
	    removeCache (new Exception ("Content-Length was not met"));
	super.finish ();
    }

    /** Remove the cachestream and the cache entry.
     * @param e the Exception that happened
     */
    protected void removeCache (Exception e) {
	super.removeCache (e);
	if (convertedFile != null)
	    convertedFile.delete ();
    }

    /** Setup this class.
     * @param prop the properties of this class.
     */
    @Override public void setup (Logger logger, SProperties prop) {
	config = prop;
	doConvert = true;
	
	String conv = prop.getProperty ("convert", STD_CONVERT);
	File f = new File (conv);
	if (!f.exists () || !f.isFile()) {
	    logger.logError (Logger.WARN, "convert -" + conv + 
			     "- not found, is your path correct?");
	    doConvert = false;
	}	
    }
}

    
