package rabbit.io;

import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;
import rabbit.util.Logger;

/** With socket channels SOTimeout does not work correctly. It leaks pipes. 
 *  This input stream takes care of that.
 */
public class NLSOInputStream extends InputStream {
    private SocketChannel sc;
    private InputStream is;
    private boolean isReady = false;
    private boolean waiting = false;
    private long waitTime;
    private NLSOHandler handler;

    public NLSOInputStream (SocketChannel sc, boolean connected, NLSOHandler handler) 
	throws IOException {
	this.sc = sc;
	this.handler = handler;
	if (!connected) {
	    waitForData ();
	}
	if (!sc.isConnected ())
	    throw new IOException ("Failed to connect: " + sc);
	this.is = sc.socket ().getInputStream ();
    }
   
    SocketChannel getSC () {
	return sc;
    }
 
    public int read () throws IOException {
	byte[] b = new byte[1];
	int i = read (b, 0, 1);
	return b[0];
    }

    public int read (byte[] b) throws IOException {
	return read (b, 0, b.length);
    }
    
    public int read (byte[] b, int off, int len) throws IOException {
	ByteBuffer bb = ByteBuffer.wrap (b, off, len);
	sc.configureBlocking (false); 
	int read = sc.read (bb);
	if (read != 0) { // -1 = EOF, > 0 got some data.
	    if (read > 0)
		sc.configureBlocking (true); 
	    return read;
	}
	isReady = false;
	waitForData ();
	if (!isReady ())
	    throw new IOException ("Timeout exceeded");
	return read (b, off, len);
    }

    private void waitForData () throws IOException {
	setWaiting (true);
	handler.addWaiting (this);
	synchronized (this) {
	    try {
		while (isWaiting ()) {
		    wait (5000);
		    if (!sc.isOpen ()) {
			setWaiting (false);
		    }
		}
	    } catch (InterruptedException e) {
		throw new IOException ("NLSO: waiting interrupted");
	    }
	}
    }
    
    private boolean isReady () {
	return isReady;
    }
    
    void setReady () {
	isReady = true;
    }

    private boolean isWaiting () {
	return waiting;
    }

    void setWaiting (boolean b) {
	waiting = b;
    }    
    
    public long skip (long n) throws IOException {
	return is.skip (n);
    }

    public int available () throws IOException {
	return 0;
    }

    public void close () throws IOException {
	is.close ();
	if (sc.isOpen ()) {
	    sc.socket ().shutdownInput ();
	    sc.socket ().shutdownOutput ();
	}
	sc.close ();
    }

    public void mark (int readlimit) {
	is.mark (readlimit);
    }

    public void reset () throws IOException {
	is.reset ();
    }

    public boolean markSupported () {
	return is.markSupported ();
    }

    long getWaitTime () {
	return waitTime;
    }
    
    void setWaitTime (long waitTime) {
	this.waitTime = waitTime;
    }
}
