package rabbit.proxy;

// $Id: RandomStream.java,v 1.12 2005/08/03 17:00:54 robo Exp $

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.RandomAccessFile;
import java.util.List;
import rabbit.cache.NCacheEntry;
import rabbit.http.GeneralHeader;
import rabbit.http.HTTPHeader;

/** This class reads data from a cachefile.
 */
public class RandomStream extends InputStream {
    private RandomAccessFile in;
    private byte[] buf;
    private int index;
    private List<Range> ranges;
    private int ri = 0;
    private String contentType;
    private String separator = "THIS_STRING_SEPARATES";
    private long totalSize;
    
    /** A single range for a resource. */
    public static class Range {
	private long start;
	private long end;
	public Range (long start, long end) {
	    this.start = start; 
	    this.end = end;
	}

	public long getStart () {
	    return start;
	}
	
	public long getEnd () {
	    return end;
	}
    }
    
    /** Create a new CacheStream from the given NCacheEntry
     * @param rh the RequestHandler for this request.
     * @param ranges the ranges that have been requested
     * @param totalSize the size to use.
     * @throws FileNotFoundException if the cachefile is not found.
     */
    public RandomStream (Connection con, 
			 Connection.RequestHandler rh, 
			 List<Range> ranges, long totalSize) 
	throws FileNotFoundException, IOException {
	NCacheEntry entry = rh.entry;
	in = new RandomAccessFile (con.getProxy ().getCache ().getEntryName (entry), "r");
	HTTPHeader oldresp = rh.getDataHook ();
	contentType = oldresp.getHeader ("Content-Type");
	this.ranges = ranges;
	this.totalSize = totalSize;
    }

    private boolean fillBuffer () throws IOException {
	index = 0;
	if (ranges.size () > 1) {
	    return fancyFill ();
	} else {
	    if (ri > 0)
		return false;
	    Range r = ranges.get (ri++);
	    int rsize = (int)(r.getEnd () - r.getStart ()) + 1;
	    buf = new byte[rsize];
	    in.seek (r.getStart ());
	    in.readFully (buf);
	    return true;
	}
    }

    private static class MultipartHeader extends GeneralHeader {
	private static final long serialVersionUID = 20050430;
	private String header;
	MultipartHeader () {
	}
	    
	public MultipartHeader (String header) {
	    this.header = header;
	}
	
	public String toString () {
	    StringBuilder ret = new StringBuilder (header);
	    ret.append (CRLF);
	    ret.append (super.toString ());
	    return ret.toString ();
	}

	public void writeExternal (ObjectOutput out) throws IOException {
	    out.writeObject (header);
	}
	
	public void readExternal (ObjectInput in) throws IOException, ClassNotFoundException {
	    header = (String)in.readObject ();
	}
    }

    private boolean fancyFill () throws IOException {
	if (ri == ranges.size ()) {
	    // CRLF should be optional according to BNF, but add it 
	    // since the rfc say it should be there.
	    buf = ("--" + separator + "--" + GeneralHeader.CRLF).getBytes ();
	    ri++;
	    return true;
	} 
	if (ri > ranges.size ())
	    return false;
	
	MultipartHeader h;
	if (ri == 0)
	    h = new MultipartHeader (GeneralHeader.CRLF + "--" + separator);
	else
	    h = new MultipartHeader ("--" + separator);
	Range r = ranges.get (ri++);
	if (contentType != null)
	    h.setHeader ("Content-Type", contentType);
	h.setHeader ("Content-Range", "bytes " + r.getStart () + "-" + r.getEnd () + "/" + totalSize);
	String hdr = h.toString ();
	int rsize = (int)(r.getEnd () - r.getStart ()) + 1;
	buf = new byte[rsize + hdr.length () + 2];
	System.arraycopy (hdr.getBytes (), 0, buf, 0, hdr.length ());	
	// currently we only handle case where we have one entry..
	in.seek (r.getStart ());
	in.readFully (buf, hdr.length (), rsize);
	buf[hdr.length () + rsize] = '\r';
	buf[hdr.length () + rsize + 1] = '\n';	
	return true;
    }
    
    public int read() throws IOException {
	if (buf == null || index >= buf.length)
	    if (!fillBuffer ())
		return -1;
	if (buf.length > index)
	    return buf[index++];
	return -1;
    }
    
    public int read(byte[] b) throws IOException {
	return read (b, 0, b.length);
    }

    public int read(byte[] b, int off, int len) throws IOException {
	if (buf == null || index >= buf.length)
	    if (!fillBuffer ())
		return -1;
	if (len > buf.length - index)
	    len = buf.length - index;
	System.arraycopy (buf, index, b, off, len);
	index += len;
	return len;
    }
    
    public void close() throws IOException {
	in.close ();
    }
}
